//   (STL )
// (c) Proteus     lawnmower-man@mail.ru

#include<stdio.h>
#include<list>

//!  
struct node_st{
	node_st(int);								
	int value;									//!<  
	std::list<node_st*>list;					//!<   
	int GetH();									//!<   
	static bool Compare(node_st*,node_st*);		//!<    	
	static node_st *Merge(node_st*,node_st*);	//!<   
	void Clear();								//!<  
	void Dump(int depth);						//!<  
	typedef std::list<node_st*>::iterator it;	//!< 
};


//!  
class BiHeap{
	std::list<node_st*>list;					//!<   
	void Insert(node_st*);						//!<     
public:
	~BiHeap();
	void Insert(int v);							//!<    
	bool isEmpty();								//!<   
	int  Pop();									//!<   
	void Clear();								//!<  
	void Dump();								//!<  
	typedef std::list<node_st*>::iterator it;	//!<  
};


//! \param v  
node_st::node_st(int v)
{
	value=v;
}


inline int node_st::GetH()
{
	return list.size();
}

bool node_st::Compare(node_st *node1,node_st *node2)
{
	return node1->GetH()==node2->GetH();
}

//!   ,      
node_st *node_st::Merge(node_st *node1,node_st *node2)
{
	node_st *node;
	if(node1->value<node2->value) node=node1,node->list.push_back(node2);
	else node=node2,node->list.push_back(node1);
	return node;
}

void node_st::Clear()
{
	for(node_st::it it=list.begin(); it!=list.end(); it++) (*it)->Clear(),delete *it;
}

//! \param depth  
void node_st::Dump(int depth)
{
	for(int n=0; n<depth; n++) putchar(' ');
	printf("[%d]\n",value);
	for(node_st::it it=list.begin(); it!=list.end(); it++) (*it)->Dump(depth+2);
}


BiHeap::~BiHeap()
{
	Clear();
}


//! \param v  
void BiHeap::Insert(int v)
{
	Insert(new node_st(v));
}

//! \param  
void BiHeap::Insert(node_st *node)
{
	BiHeap::it it;
	for(;;) {
		for(it=list.begin(); it!=list.end(); it++) if(node_st::Compare(node,*it)) break;
		if(it==list.end()) break;
		node=node_st::Merge(node,*it);
		list.erase(it);
	}
	list.push_back(node);
}

int BiHeap::Pop()
{
	int min;
	node_st::it it1,it2;
	//      
	min=list.front()->value+1;
	for(it1=list.begin();it1!=list.end();it1++) if(min>=(*it1)->value) it2=it1,min=(*it1)->value;
	//    
	node_st *node=*it2; list.erase(it2);
	for(it1=node->list.begin(); it1!=node->list.end(); it1++) Insert(*it1);
	delete node;
	return min;
}

bool BiHeap::isEmpty()
{
	return list.empty();
}

void BiHeap::Dump()
{
	printf("[dump]\n");
	for(node_st::it it=list.begin(); it!=list.end(); it++) (*it)->Dump(0);
}

void BiHeap::Clear()
{
	for(node_st::it it=list.begin(); it!=list.end(); it++) (*it)->Clear(),delete *it;
}



#define SIZE 30

void main()
{
	BiHeap heap;
	for(int n=0; n<SIZE; n++) heap.Insert(rand()%30);
	heap.Dump();
	while(!heap.isEmpty()) printf("(%d)",heap.Pop());
	putchar('\n');
	printf("[press enter]\n");
	getchar();
}
